#!/usr/bin/perl
#
# Package for operating a Amplus 03127 LED display
#
# Created by Vegard Hanssen @ Xait AS

package Amplus::03127;

use strict;
use Device::SerialPort qw( :PARAM :STAT 0.07);
our $VERSION = '0.4';

sub new {
	my $class = shift;

	my $self = {};
	bless $self, $class;
	$self->_Init();
	return $self;
}

# Send the data in the object to the led
sub Send {
	my ($self) = @_;

	if ( defined( $self->{'BLOCKS'} ) ) {
		return $self->_SendGraphics();
	}
	else {
		return $self->_SendString( $self->_getString() );
	}
}

# Set a page up for run
# In: page, 1-26
# Out: 1 = success
#      0 = failed
sub RunPage {
	my ( $self, $page ) = @_;

	if ( $page > 0 && $page <= 26 ) {
		$page = chr( ord('A') + $page - 1 );
	}
	else {
		return 0;
	}
	return $self->_SendString( $self->_AddStandardOnText("<RP$page>") );
}

# Set the ID on the LED
# You must connect only one LED when you do this
# In: id, 0-255
# Out: 1 = success
#      0 = failed
sub SetID {
	my ( $self, $id ) = @_;

	if ( $id >= 0 && $id <= 255 ) {
		$id = sprintf( "%02x", $id );
	}
	else {
		return 0;
	}
	return $self->_SendString("<ID><$id><E>");
}

# Set time on led to computer clock
# Out: 1 = success
#      0 = failed
sub SetTime {
	my ($self) = @_;

	my ( $sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst ) =
	  localtime;
	my $str = "<SC>";

	# We need the 2 last digit in the year, perl returns years since 1900
	$year += 1900;
	$year = substr( $year, 2, 2 );
	my $time = sprintf( "%02d%02d%02d%02d%02d%02d%02d",
		$year, $wday + 1, $mon + 1, $mday, $hour, $min, $sec );
	return $self->_SendString( $self->_AddStandardOnText( $str . $time ) );
}

# Set the brightness on the LED
# In: brightness
#       1 = 100%
#       2 = 75%
#       3 = 50%
#       4 = 25%
# Out: 1 = success
#      0 = failed
sub SetBrightness {
	my ( $self, $value ) = @_;

	if ( $value >= 1 && $value <= 4 ) {

		# Internly A=100, B=75, C=50, D=25
		$value = chr( ord('A') + $value - 1 );
	}
	else {
		return 0;
	}
	return $self->_SendString( $self->_AddStandardOnText("<B$value>") );
}

# Initializing object with standard values for the led
sub _Init {
	my ($self) = @_;

	$self->TTY('/dev/ttyS0');
	$self->ID('00');
	$self->Line(1);
	$self->Page(1);
	$self->BlendIn(5);
	$self->BlendOut(5);
	$self->ScrollSpeed(1);
	$self->DisplayTime(2);
	$self->_getConnection();
}

# Sending a string to the led
#
# In: $str = formatted string for the led
#
# Out: 1 = success
#      0 = failed
sub _SendString {
	my ( $self, $str ) = @_;

	# Get the serial-object
	my $po = $self->{'PO'};

	# Seems the led want the data in 8 blocks
	# (or alternating 16/8 from block 4)
	my $delta = 8;
	for ( my $i = 0 ; $i < length($str) ; $i += $delta ) {
		my $substr = substr( $str, $i, $delta );
		$po->write($substr);
	}
	my $result = 0;

	# Read the response from led
	# This should be ACK, or else something is wrong
	my $counter = 0;
	while (1) {
		my ( $count, $str ) = $po->read(20);
		if ( $count > 0 ) {
			if ( $str eq "ACK" ) {
				$result = 1;
			}
			else {
				$result = 0;
			}
			last;
		}
		sleep(1);

		# Only wait for 10 seconds
		if ( $counter++ > 10 ) {
			last;
		}
	}
	return $result;
}

# Generates the string to send
#
# Return: Formatted string to send
sub _getString {
	my ($self) = @_;

	my $str = "";
	$str .= "<L" . $self->Line() . ">";
	$str .= "<P" . $self->Page() . ">";
	$str .= "<F" . $self->BlendIn() . ">";
	if ( $self->Song() ) {
		$str .= "<M" . $self->Song() . ">";
	}
	if ( $self->ScrollSpeed() ) {
		$str .= "<" . $self->ScrollSpeed() . ">";
	}
	$str .= "<W" . $self->DisplayTime() . ">";
	$str .= "<F" . $self->BlendOut() . ">";
	if ( $self->FontSize() ) {
		$str .= "<A" . $self->FontSize() . ">";
	}
	$str .= $self->Text();
	$str = $self->_AddStandardOnText($str);
	return $str;
}

# Adds standard codes on text. This
# includes the ID-tag, checksum and the
# ending E-tag.
# The checksum is calculated on the string
# without the ID- and E-tag
sub _AddStandardOnText {
	my ( $self, $str ) = @_;

	$str .= $self->_calcCheckSum($str);
	$str = "<ID" . $self->ID() . ">" . $str;
	$str .= "<E>";
	return $str;
}

# Send pages which will be run on the led
# @pages = array of pages, 1 to 26
#
# Out: 1 = success
#      0 = failed
sub SendPagesToRun {
	my ( $self, @pages ) = @_;

	my $str = "<TA>00010100009912302359";
	for my $page (@pages) {

		# If page is out of bound, return false
		if ( $page < 1 || $page > 26 ) {
			return 0;
		}

		# Pages are named A-Z in the led
		$str .= chr( ord('A') + $page - 1 );
	}
	$str = $self->_AddStandardOnText($str);
	return $self->_SendString($str);
}

# Calculates the checksum for the message
#
# In: The string
# Out: The checksum
sub _calcCheckSum {
	my ( $self, $str ) = @_;

	my @sum = ();

	# Loop through all the characters
	for my $pos1 ( 0 .. length($str) - 1 ) {
		my $char = substr( $str, $pos1, 1 );

		# Convert the character to 8 bits
		my $bin = $self->_charToBin($char);

		# Sum the bits position
		for my $pos2 ( 0 .. length($bin) - 1 ) {
			$sum[$pos2] += substr( $bin, $pos2, 1 );
		}
	}

	my $sumstr = "";

	# Loop through the sums of the 8 positions
	for ( 0 .. 7 ) {
		my $tmp = sprintf( "%b", $sum[$_] );

		# Add the first bit
		$sumstr .= substr( $tmp, length($tmp) - 1, 1 );
	}

	# Convert the new byte to an integer
	my $int = oct("0b$sumstr");

	# Return the hex of the integer
	return uc sprintf( "%02x", $int );
}

# Converts a charater to 8 bits
# In: Character
# Out: 8 bits as a string
sub _charToBin {
	my ( $self, $char ) = @_;

	my $bin = sprintf( "%b", ord($char) );
	while ( length($bin) < 8 ) {
		$bin = "0" . $bin;
	}
	return $bin;
}

# get/set method for TTY
# In: tty, /dev/ttyS0
# Out: tty
sub TTY {
	my ( $self, $value ) = @_;

	if ($value) {
		$self->{'TTY'} = $value;
	}
	return $self->{'TTY'};
}

# get/set method for ID of led
# In: ID, 0-255
# Out: ID, 00-FF
sub ID {
	my ( $self, $value ) = @_;

	if ($value) {

		# led is using hex values in uppercase
		my $hex = uc sprintf( "%02x", $value );
		$self->{'ID'} = $value;
	}
	return $self->{'ID'};
}

# get/set method for Page of led
# In: Page, 1-26
# Out: Page, A-Z
sub Page {
	my ( $self, $value ) = @_;

	if ( $value ne "" ) {
		if ( $value >= 1 && $value <= 26 ) {
			$self->{'PAGE'} = chr( ord('A') + $value - 1 );
		}
		else {
			return 0;
		}
	}
	return $self->{'PAGE'};
}

# get/set method for Line of led
# This is disabled because the
# led only supports 1 line.
sub Line {
	my ( $self, $value ) = @_;

	if ($value) {
		$self->{'LINE'} = 1;    #$value;
	}
	return $self->{'LINE'};
}

# get/set method for blendin effect
# In: effect, 0-14, 0=random
# Out: effect, A-N,P
sub BlendIn {
	my ( $self, $value ) = @_;

	return $self->_Blend( $value, 'BLENDIN' );
}

# get/set method for blendout effect
# In: effect, 0-14, 0=random
# Out: effect, A-N,P
sub BlendOut {
	my ( $self, $value ) = @_;

	return $self->_Blend( $value, 'BLENDOUT' );
}

# In: effect, 0-14, 0=random
#     type, BLENDIN, BLENDOUT
# Out: effect, A-N,P
sub _Blend {
	my ( $self, $value, $type ) = @_;

	if ( $value ne "" ) {
		if ( $value == 0 ) {
			$value = 'P';
		}
		elsif ( $value >= 1 && $value <= 14 ) {
			$value = chr( ord('A') + $value - 1 );
		}
		else {
			return 0;
		}
		$self->{$type} = $value;
	}
	return $self->{$type};

}

# get/set method for scrollspeed
# In: speed, 1-4 1=slow - 4=fast
#     blink, 1=on, 0=off (default)
# Out: speedstring
sub ScrollSpeed {
	my ( $self, $value, $blink ) = @_;

	my %values = (
		'4'  => 'MA',
		'3'  => 'MQ',
		'2'  => 'Ma',
		'1'  => 'Mq',
		'41' => 'MB',
		'31' => 'MR',
		'21' => 'Mb',
		'11' => 'Mr',
	);
	if ($value) {
		$self->{'SCROLLSPEED'} = $values{"$value$blink"};
	}
	return $self->{'SCROLLSPEED'};
}

# get/set method for fontsize
# In: size, 1-4
# Out: A,B,C,F
sub FontSize {
	my ( $self, $value ) = @_;

	my %values = (
		'1' => 'C',
		'2' => 'A',
		'3' => 'B',
		'4' => 'F',
	);
	if ($value) {
		$self->{'FONTSIZE'} = $values{$value};
	}
	return $self->{'FONTSIZE'};
}

# get/set method for displaytime
# In: 0-25, 0=0.5s, 1=1s, 2=2s...
# Out: A-Z
sub DisplayTime {
	my ( $self, $value ) = @_;

	if ( $value ne "" ) {
		if ( $value >= 0 && $value <= 25 ) {
			$self->{'DISPLAYTIME'} = chr( ord('A') + $value );
		}
		else {
			return 0;
		}
	}
	return $self->{'DISPLAYTIME'};
}

# Set the text-attribute to date
# Out: 1 = success
#      0 = failed
sub ShowDate {
	my ($self) = @_;

	return $self->_SendString( $self->_AddStandardOnText('<KD>') );
}

# Set the text-attribute to datetime
# Out: 1 = success
#      0 = failed
sub ShowDateTime {
	my ($self) = @_;

	return $self->_SendString( $self->_AddStandardOnText("<KD> <KT>") );
}

# Set the text-attribute to time
# Out: 1 = success
#      0 = failed
sub ShowTime {
	my ($self) = @_;

	return $self->_SendString( $self->_AddStandardOnText("<KT>") );
}

# get/set method for the text
# In: text
# Out: text
sub Text {
	my ( $self, $value ) = @_;

	if ($value) {
		$self->{'TEXT'} = $value;
	}
	return $self->{'TEXT'};
}

# get/set method for the song
# In: 1-3
# Out: C-E
sub Song {
	my ( $self, $value ) = @_;

	if ($value) {
		if ( $value >= 1 && $value <= 3 ) {
			$self->{'SONG'} = chr( ord('B') + $value );
		}
		else {
			return 0;
		}
	}
	return $self->{'SONG'};
}

# Start the serialport
# Out: Device::SerialPort-object
sub _getConnection {
	my ($self) = @_;

	if ( !$self->{'PO'} ) {
		my $po = new Device::SerialPort( $self->TTY(), 1 );
		$po->handshake("xoff");
		$po->baudrate(9600);
		$po->parity("none");
		$po->databits(8);
		$po->stopbits(1);
		$po->restart;

		$self->{'PO'} = $po;
	}
	return $self->{'PO'};    
}

# Init graphic objects
# In: array of the numbers 1-16
# Out: 1 = success
#      0 = failed
sub InitGraphics {
	my ( $self, @blocks ) = @_;    # one block = 32 leds

	for my $block (@blocks) {
		if ( $block >= 1 && $block <= 16 ) {
			push @{ $self->{'BLOCKS'} }, $block;
		}
		else {
			return 0;
		}
	}
	return 1;
}

# Flush the graphic object
sub FlushGraphics {
	my ($self) = @_;

	undef $self->{'BLOCKS'};
}

# Draw a point in the graphic
# 0,0 is upper left
sub DrawPoint {
	my ( $self, $x, $y ) = @_;

	$self->{'LEDS'}{$x}{$y} = 1;
}

# Sends the graphic to the led
# Out: 1 = success
#      0 = failed
sub _SendGraphics {
	my ($self) = @_;

	# First we need to generate the controll-block
	# This is the same as for text, only with
	# <GCx>s for the blocks as text
	my @blocks = @{ $self->{'BLOCKS'} };
	my $text   = "";
	for ( my $block = 0 ; $block < $#blocks + 1 ; $block++ ) {
		$text .= "<GC" . $blocks[$block] . ">";
	}
	$self->Text($text);
	if ( !$self->RunString() ) {
		return 0;
	}

	# Now we send the graphic blocks
	for ( my $block = 0 ; $block < $#blocks + 1 ; $block++ ) {
		$text = "<GC" . $blocks[$block] . ">";
		$text .= $self->_MakeBlockGraphic($block);
		$text = $self->_AddStandardOnText($text);
		if ( !$self->_SendString($text) ) {
			return 0;
		}
	}
	return 1;
}

# Generate a graphic block
# Out: The graphic
sub _MakeBlockGraphic {
	my ( $self, $block ) = @_;

	my $result = "";
	my $delta  = $block * 32;

	# Each block is 4x8x8 pixels
	# xxxx xxxx
	# xxxx xxxx
	# ....
	#           xxxx xxxx
	#           xxxx xxxx
	#           ...
	for ( my $col = 0 ; $col < 4 ; $col++ ) {
		for ( my $y = 0 ; $y < 8 ; $y++ ) {
			my $word = "";
			for ( my $x = 0 ; $x < 8 ; $x++ ) {
				if ( $self->{'LEDS'}{ $x + $delta + ( $col * 8 ) }{$y} == 1 ) {

					# A light is defined by 2-bits, 10
					$word .= "10";
				}
				else {

					# A black is defined by 2-bits, 00
					$word .= "00";
				}
			}

			# Pack the first 8 pixels (16 bits string) into 2 bytes
			$result .= pack( "n", oct("0b$word") );
		}
	}
	return $result;
}

42;

__END__

=head1 NAME

Amplus::03127 - LED display

=head1 SYNOPSIS

 use Amplus::03127;

 $led = new Amplus::03127;
 $led->Text("Hello World");
 $led->Send();

=head1 DESCRIPTION

This module helps you program the Amplus 03127 LED display.
You can set both text and graphics with it.

=head1 USAGE

 $led = new Amplus::03127;

When you create an object, the module will setup default values
for it.

=head2 METHODS

=over 4

=item I<PACKAGE>->new

Returns a newly created C<Amplus::03127> object. The object will be
populated with standard values.

    TTY=/dev/ttyS0
    ID=0
    Line=1
    Page=1
    BlendIn=5
    BlendOut5
    ScrollSpeed=1
    DisplayTime=2

=item I<$OBJ>->Send

Sends the data to the LED. Data can be graphics or text.

Returns 1 if success, 0 if failed

=item I<$OBJ>->RunPage(C<Page>)

Sends which page to be run right now on the LED. Page can be 1-26.

Returns 1 if success, 0 if failed

=item I<$OBJ>->SendPagesToRun(C<@pages>)

Sends which pages to run. This will make the led alternate between these pages.
You must populate the pages with something, or else you get a no data for
that page.

	$led->SendPagesToRun(qw(1 3 5));

Returns 1 if success, 0 if failed

=item I<$OBJ>->SetID(C<ID>)

Set the ID on the LED. You should only connect one LED to the serialport
before running this. ID can be 0-255.

Returns 1 if success, 0 if failed

=item I<$OBJ>->SetTime()

Set the time on the LED to computers time.

Returns 1 if success, 0 if failed

=item I<$OBJ>->SetBrightness(C<Brightness>)

Brightness is
    1 = 100%
    2 = 75%
    3 = 50%
    4 = 25%

Returns 1 if success, 0 if failed

=item I<$OBJ>->TTY([C<TTY>])

Set the serialport to use. Standard is /dev/ttyS0.

Will return the value set for the TTY

=item I<$OBJ>->ID([C<ID>])

Set the ID to work with. This is not the same as C<SetID> which set the ID on the LED.
You must always set the ID on the object to send the command to correct LED. Default
this is set to 0. You can use 0-255 as ID

Returns the value of ID as hex 00-FF

=item I<$OBJ>->Page([C<Page>])

Set the page to work with. Default value is 1. Possible values are 1-26.
(Internally these are A-Z).

Returns the value set for the page as A-Z

=item I<$OBJ>->Line([C<Line>])

Set the line to work with. The 03127 LED only support 1 line, so setting it to other
than 1 will set it to 1.

Returns the value set for the line

=item I<$OBJ>->BlendIn([C<effect>])

Set the C<BlendIn>-effect for the text. Possible values are 0-14

	0 = Random
	1 = Immediate
	2 = Xopen
	3 = Curtain Up
	4 = Curtain Down
	5 = Scroll Left
	6 = Scroll Right
	7 = Vopen
	8 = Vclose
	9 = Scroll Up
	10 = Scrull Down
	11 = Hold
	12 = Snow
	13 = Twinkle
	14 = Block Move	 
	
Returns the value for effect, A-N,P

=item I<$OBJ>->BlendOut([C<effect>])

Same as C<BlendIn>

=item I<$OBJ>->ScrollSpeed([C<speed>], [C<blink>])

Set the C<speed> of the scroll. Possible values 1-4, where
1=slow, 4=fast.

C<blink> will set the text to blink, must be value 1.

Returns the command for scrollspeed.

	MA=speed 4, MQ=speed 3, Ma=speed 2, Mq=speed 1
	MB=speed 4 blink, MR=speed 3 blink, Mb=speed 2 blink, Mr=speed 1 blink
	
=item I<$OBJ>->FontSize([C<fontsize>])

Set the size of the font. This is not very useful on the 03127.
Possible values 1-4

Returns A,B,C,F

=item I<$OBJ>->DisplayTime([C<time>])

Set the time to display the text. Possible values 0-25

0=0.5s, 1=1s, 2=2s..

Returns A-Z

=item I<$OBJ>->ShowDate()

Set the text on the LED to the date. You have to set the page etc before
calling this command.

=item I<$OBJ>->ShowDateTime()

Same as C<ShowDate()>, only showing date and time

=item I<$OBJ>->ShowTime()

Same as C<ShowDate()>, only showing time

=item I<$OBJ>->Text([C<text>])

Set the text on the object. You must call C<Send> to send it to the LED

=item I<$OBJ>->Song([C<song>])

Set the song to play on the LED. (extremly annoing).
Possible values are 1-3.

Returns the value for song, C-E

=item I<$OBJ>->InitGraphics(C<@blocks>)

The LED as 16 blocks of graphics which can be combined in different pages.
You can have 4 blocks on each page. So you can setup block 1,4,6 on page 1,
2,5,7 on page 2 etc. Each block is 7x32 pixels. (Actually it's sent as 8x32,
but the LED will not display the 8th line.)

$led->InitGraphics(qw(1 4 6));

This will tell the object that you want to render graphic on 3 blocks, meaning
7x96 pixels. The blocks will be stored as 1,4 and 6 in the LED. If you have
setup other pages with those block, those pages will also be altered.

Returns 1 if you gave it correct data, 0 if not

=item I<$OBJ>->FlushGraphics()

Will flush the graphic from the object, so you can send text again.
This will not remove any graphics from the LED.

=item I<$OBJ>->DrawPoint(C<$x>, C<$y>)

This will draw a point on the graphic-object at the x,y coordinate.
0,0 is up to the left. After you have set all the points, you call
$led->Send() to send it to the LED.

=back

=head1 EXAMPLES

	$led = new Amplus::03127;
	
	# Send a text to page 1
	$led->Text("Hello World");
	$led->Send();
	
	# Send the same text to page 2
	$led->Page(2);
	$led->Send();
	
	# Set a new text on page 3
	$led->Page(3);
	$led->Text("A new text");
	$led->Send();
	
	# Run page 1 and 3
	$led->SendPagesToRun(qw(1 3));
	
	# Only run page 2
	$led->RunPage(2);
	
	# Change speed on page 2 text
	$led->ScrollSpeed(4);
	$led->Send();
	
	# Change Blendin on page 2
	$led->BlendIn(3);
	$led->Send();
	
	# Work with another LED with ID 2
	$led->ID(2);
	
	# Create a graphic object to render on
	# Will use 2 blocks, 1 and 9
	$led->ID(0);
	$led->Page(1);
	$led->InitGraphics(qw(1 9));
	$led->DrawPoint(0,0);
	$led->DrawPoint(7,0);
	$led->DrawPoint(0,63);
	$led->DrawPoint(7,63);
	$led->Send();
	
	# Create another graphic object to render
	# on page 2
	$led2 = new Amplus::03127();
	$led2->Page(2);
	$led2->InitGraphics(qw(2 3));
	$led2->DrawPoint(1,1);
	$led2->DrawPoint(6,1);
	$led2->DrawPoint(1,62);
	$led2->DrawPoint(6,62);
	$led2->Send();
	
	# Animate between the 2 pages
	$led->BlendIn(1); # Immediate
	$led2->BlendIn(1);
	$led->BlendOut(1);
	$led2->BlendOut(1);
	$led->DisplayTime(0); # 0.5s
	$led2->DisplayTime(0);
	$led->Send();
	$led2->Send();
	$led->SendPagesToRun(qw(1 2));
		
	# Draw some text again
	$led->FlushGraphics();
	$led->Text("Some text");
	$led->Send();

=head1 TODO

	Create graphics out of GD objects, including animated gifs.
	
	Add Scheduling. The LED has 5 slots for scheduling pages run
	run between two datetimes.

=head1 AUTHOR

The C<Amplus::03127> module is written by Vegard Hanssen <Vegard@menneske.no>.
Credits go to Xait AS (www.xait.no) for letting me work on it. If you should
use this module, I would be very happy to hear about it.

=head1 COPYRIGHT

 Copyright 2007, Vegard Hanssen

This library is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

=cut
